import math
import csv
import os.path


def spec2cms(filename: str, sldir: float, st_depth: float, gr_depth: float) -> None:
    """
    filename (str):   Filename of WIS 2d spectra
    sldir (float):    CMS-Wave x-axis orientation in Cartesian coordinates (degrees)
    st_depth (float): Station depth (m)
    gr_depth (float): Grid seaward boundary depth (m)
    """
    # For each line of .spe2d file, process and output to .txt. file
    if os.path.exists(filename):
        f1 = open(filename, 'r')
    else:
        return

    # Remove leading ./ or .\ off the filename
    name = filename.split('/')[-1]

    prefix = name.split('.')[0]
    f2 = open(f"{prefix}.eng", 'w')
    f3 = open(f"{prefix}.txt", 'w')
    f4 = open(f"{prefix}.dat", 'w')
    f5 = open(f"{prefix}.out", 'w')
    f6 = open(f"{prefix}.asc", 'w')

    # A few constants for use later
    con = 0.01745329
    tide = 0.
    isid = 0
    nsi = 20
    minus1 = -1
    a999 = 999.

    pi = math.pi
    rad = pi/180.
    power2 = min(math.log(st_depth/gr_depth), 30.)

    line = f1.readline()
    args = line.split()
    nf = int(args[4])
    nang = int(args[5])
    nang2 = int(nang/2)
    kang = nang2 - 1    # Changed to depend on the number of angles read in from the 2d spectra (i.e., Alaska)
    bang = 360. / nang  # Changed to depend on the number of angles read in from the 2d spectra
    deltc = bang*rad    # Moved down because of the dependence on 'bang'

    # Skip remaining two lines in the header
    f1.readline()
    f1.readline()

    # Read frequencies into list
    hz = []
    for freq in range(nf):
        line = f1.readline()
        args = line.split()
        hz.append(float(args[0]))
    f1.close()

    cs1 = []
    cs2 = []
    sn1 = []
    sn2 = []
    for k in range(nang):
        dd1 = (bang * float(k+1) - 90.) * con
        cs1.append(math.cos(dd1))
        sn1.append(math.sin(dd1))
        dd2 = (bang * float(k+1) - 2.5) * con
        cs2.append(math.cos(dd2))
        sn2.append(math.sin(dd2))

    deltf = [0]
    for j in range(1, nf - 1):
        deltf.append(round((hz[j+1]-hz[j-1])/2., 5))
    deltf[0] = deltf[1]
    deltf.append(deltf[-1])

    # Write number of frequencies, number of direction bins, and list of frequencies to .asc file
    f6.write(f"{nf:>3}{nang:>3}\n")
    hz_list = [f"{item:8.5f}" for item in hz]
    strlen = len(hz_list)
    for i in range(int(strlen/10)):  # This will output only 10 values per line and continue on the next.
        inc = (i+1)*10-1
        hz_list[inc] = hz_list[inc]+'\n'
    if len(hz_list) % 10 != 0:
        hz_list[-1] = hz_list[-1] + '\n'
    f6.writelines(hz_list)

    if sldir < 0:
        sldir = sldir+360.
    if sldir >= 360.0:
        sldir = sldir-360.
    ishift = int(sldir/bang)

    # Write main header to .txt file
    f3.write('SPECTRAL TABLE\n')
    f3.write('      METHOD      OPTION        TIME           INDEX    ANGLE(1)       HS(1)       TP(1)'
             '    GAMMA(1)    ANGLE(2)       HS(2)       TP(2)    GAMMA(2)        WIND       FETCH       NN(1)'
             '       NN(2)   STDDEV(1)   STDDEV(2)       DEPTH\n')

    # Write number of frequencies, number of direction bins, and list of frequencies to .eng file
    f2.write(f"{nf:>3}{kang:>3}\n")
    f2.writelines(hz_list)

    with open(filename, 'r') as f1:
        reader = csv.reader(f1, delimiter=' ')  # Easier to read in table-type values with CSV functions.
        for line in reader:
            if len(line) == 0:
                try:
                    line = reader.__next__()
                except StopIteration:
                    break
            line = [i for i in line if i]
            for i in range(line.count('\t29\t')):
                line.remove('\t29\t')
            iyr = int(line[1][0:4])
            imo = int(line[1][4:6])
            ida = int(line[1][6:8])
            ihr = int(line[1][8:10])
            itime = f"{iyr:4d}{imo:02d}{ida:02d}{ihr:02d}00"

            line = reader.__next__()
            line = [i for i in line if i]
            ws = float(line[0])
            wd = float(line[1])

            line = reader.__next__()
            line = [i for i in line if i]
            hsi = float(line[0])
            tdi = float(line[1])
            tai = float(line[3])
            jpeak = nf-1

            # hz = []
            ee = []
            for freq in range(nf):
                line = reader.__next__()
                line = [i for i in line if i]
                line = [float(i) for i in line]
                # hz.append(line[0])
                ee.append(line[1:])  # Can skip over the first value, hz, which we already have

            sumx = 0.
            sumy = 0.
            for j in range(nf):
                for k in range(nang):
                    sumx = sumx + (ee[j][k]*cs2[k]*deltf[j])
                    sumy = sumy + (ee[j][k]*sn2[k]*deltf[j])
            hdii = math.atan2(sumx, sumy)/rad
            hdii = 270.-hdii
            if hdii < 0.:
                hdii = hdii+360.
            if hdii >= 360.:
                hdii = hdii-360.

            # Write values to .out file
            f5.write(f"{isid:4d}{iyr:5d}{imo:3d}{ida:3d}{ihr:3d}{hsi:7.2f}{tdi:6.1f}{tai:6.1f}{hdii:5.0f}."
                     f"{ws:6.1f}{wd:5.0f}.\n")

            # Write record header and spectral values to .asc file
            f6.write(f"{itime:>12}{ws:8.1f}{wd:10.0f}.{1./tdi:10.5f}{float(isid):8.2f}\n")
            ee_list = []
            for record in ee:
                for item in record:
                    ee_list.append(f"{item:11.5f}")
            strlen = len(ee_list)
            for i in range(int(strlen/7)):  # This will output only 7 values per line and continue on the next.
                inc = (i+1)*7-1
                ee_list[inc] = ee_list[inc]+'\n'
            if len(ee_list) % 7 != 0:
                ee_list[-1] = ee_list[-1] + '\n'
            f6.writelines(ee_list)

            big = 0.
            asum1 = 0.
            asum2 = 0.
            sumx = 0.
            sumy = 0.

            es = [[0. for _ in range(nang)] for _ in range(nf)]
            for k in range(1, kang+1):  # Because of the way Python does its ranges, shift up one number
                kshift = nang2 - k - ishift
                if kshift > nang:
                    kshift = kshift-nang
                if kshift <= 0:
                    kshift = kshift+nang
                if kshift > nang:
                    kshift = kshift-nang
                if kshift <= 0:
                    kshift = kshift+nang
                for j in range(nf):
                    es[j][k-1] = ee[j][kshift-1]  # Shift access value one spot lower for arrays

            sumd = 0.
            for j in range(nf):
                sump = 0.
                for k in range(nang):
                    ee[j][k] = es[j][k]
                    sump = sump + ee[j][k]
                sumd = sumd + (sump*deltf[j])
            if st_depth != gr_depth:
                for k in range(kang):
                    for j in range(nf):
                        val = ee[j][k]*(cs1[k]**power2)  # cs1[k] must be a positive number since 0 < power2 < 1
                        ee[j][k] = float(val)

            e1 = []
            for j in range(nf):
                e1.append(0.)
                for k in range(nang):
                    e1[j] = e1[j]+ee[j][k]
                    sumx = sumx + ee[j][k]*cs1[k]
                    sumy = sumy + ee[j][k]*sn1[k]
                e1[j] = e1[j]*deltf[j]
                asum1 = asum1 + e1[j]
                asum2 = asum2 + e1[j]*hz[j]
                if e1[j] > big:
                    big = e1[j]
                    jpeak = j

            hsii = 4.*math.sqrt(asum1*deltc)  # hsii is the transformed half-plane wave ht
            asum2 = 0.0000001 if asum2 == 0.0 else asum2  # avoid divide by zero
            taii = asum1/asum2

            # Write record header to .eng file
            f2.write(f"{itime:>12}{ws:8.1f}{wd:11f}{hz[jpeak]:10.5f}{tide:8.2f}\n")

            tdii = 1/hz[jpeak]
            gami = 3. + math.tanh(.1/hz[jpeak])
            wmd = math.atan2(sumy, sumx) / rad
            hdjj = wmd

            wmd = -sldir - 90. - wmd
            if wmd < 0.:
                wmd = wmd + 360.
            if wmd >= 360.:
                wmd = wmd - 360.

            cc = 1.0
            hsii = hsii*math.sqrt(cc)

            # Write record spectral values to .eng file
            eecc_list = []
            for j in range(nf):
                for k in range(kang):
                    val = ee[j][k]
                    eecc_list.append(f"{val:11.5f}")
            strlen = len(eecc_list)
            for i in range(int(strlen / 7)):  # This will output 7 values per line and continue on the next.
                inc = (i+1)*7-1
                eecc_list[inc] = eecc_list[inc] + '\n'
            if len(eecc_list) % 7 != 0:
                eecc_list[-1] = eecc_list[-1] + '\n'
            f2.writelines(eecc_list)

            # Write record values to .txt file
            f3.write(f"{isid:12d}{minus1:12d}{a999:12.4f}{itime:>16}{hdjj:12.4f}{hsii:12.4f}{tdii:12.4f}{gami:12.4f}"
                     f"{a999:12.4f}{a999:12.4f}{a999:12.4f}{a999:12.4f}{ws:12.4f}{a999:12.4f}{nsi:12d}{isid:12d}"
                     f"{a999:12.4f}{a999:12.4f}{st_depth:12.4f}\n")

            # Write record values to .dat file
            f4.write(f"{isid:4d}{iyr:5d}{imo:3d}{ida:3d}{ihr:3d}{hsii:7.2f}{tdii:6.1f}{taii:6.1f}{wmd:5.0f}."
                     f"{ws:6.1f}{wd:5.0f}.\n")
    f2.close()
    f3.close()
    f4.close()
    f5.close()
    f6.close()


if __name__ == "__main__":
    # For testing - Alaska station
    fname: str = 'ST82251-2d_spectra-20240416T12_45.spe2d'
    orient: float = 93.0
    station_depth: float = 28.0
    grid_depth: float = 20.1
    spec2cms(fname, orient, station_depth, grid_depth)
